@@ -17,8 +17,7 @@ import sublime_api
 from typing import Optional, TYPE_CHECKING
 
 if TYPE_CHECKING:
-    from sublime_types import DIP, Vector, Point, Value, CommandArgs, Kind, Event, CompletionValue
-
+    from sublime_types import Value, Event
 
 api_ready = False
 
@@ -652,9 +651,9 @@ def check_all_view_event_listeners():
             check_view_event_listeners(v)
 
 
-def detach_view(view):
-    if view.view_id in view_event_listeners:
-        del view_event_listeners[view.view_id]
+def detach_view(view_id):
+    if view_id in view_event_listeners:
+        del view_event_listeners[view_id]
 
     # A view has closed, which implies 'is_primary' may have changed, so see if
     # any of the ViewEventListener classes need to be created.
@@ -919,12 +918,7 @@ def on_pre_close(view_id):
 
 
 def on_close(view_id):
-    v = sublime.View(view_id)
-    for callback in vel_callbacks(v, 'on_close'):
-        callback()
-    detach_view(v)
-    for callback in el_callbacks('on_close'):
-        callback(v)
+    run_view_callbacks('on_close', view_id)
 
 
 def on_pre_save(view_id):
@@ -1000,91 +994,34 @@ def on_query_context(view_id, key, operator, operand, match_all):
     return False
 
 
-def normalise_completion(c):
-    def split_trigger(trigger):
-        idx = trigger.find("\t")
-        if idx < 0:
-            return (trigger, "")
-        else:
-            return (trigger[0:idx], trigger[idx + 1:])
-
-    if not isinstance(c, sublime.CompletionItem):
-        if isinstance(c, str):
-            trigger, annotation = split_trigger(c)
-            c = sublime.CompletionItem(trigger, annotation)
-        elif len(c) == 1:
-            trigger, annotation = split_trigger(c[0])
-            c = sublime.CompletionItem(trigger, annotation)
-        elif len(c) == 2:
-            trigger, annotation = split_trigger(c[0])
-            c = sublime.CompletionItem.snippet_completion(
-                trigger,
-                c[1],
-                annotation,
-                kind=sublime.KIND_AMBIGUOUS
-            )
-        elif len(c) == 3:
-            trigger, annotation = split_trigger(c[0])
-            c = sublime.CompletionItem.snippet_completion(
-                trigger,
-                c[2],
-                annotation,
-                kind=sublime.KIND_AMBIGUOUS
-            )
-        else:
-            c = sublime.CompletionItem("")
-
-    kind, kind_letter, kind_name = c.kind
-
-    letter = 0
-    if isinstance(kind_letter, str) and kind_letter != '':
-        letter = ord(kind_letter)
-    return (c.trigger, c.annotation, c.details, c.completion, kind_name, letter, c.completion_format, c.flags, kind)
-
-
-class MultiCompletionList:
-    def __init__(self, num_completion_lists, view_id, req_id):
-        self.remaining_calls = num_completion_lists
-        self.view_id = view_id
-        self.req_id = req_id
-        self.completions = []
-        self.flags = 0
-
-    def completions_ready(self, completions, flags):
-        self.completions += [normalise_completion(c) for c in completions]
-        self.flags |= flags
-        self.remaining_calls -= 1
-
-        if self.remaining_calls == 0:
-            sublime_api.view_set_completions(
-                self.view_id, self.req_id, (self.completions, self.flags))
+def split_trigger(trigger):
+    idx = trigger.find("\t")
+    if idx < 0:
+        return (trigger, "")
+    else:
+        return (trigger[0:idx], trigger[idx + 1:])
 
 
 def on_query_completions(view_id, req_id, prefix, locations):
     v = sublime.View(view_id)
 
-    completion_lists = []
+    mlist = sublime_api.MultiCompletionList(view_id, req_id)
 
     def norm_res(res):
         if isinstance(res, tuple):
-            completion_lists.append(sublime.CompletionList(res[0], flags=res[1]))
+            return sublime.CompletionList(res[0], flags=res[1])
         elif isinstance(res, list):
-            completion_lists.append(sublime.CompletionList(res))
+            return sublime.CompletionList(res)
         elif isinstance(res, sublime.CompletionList):
-            completion_lists.append(res)
+            return res
 
     for callback in el_callbacks('on_query_completions'):
-        norm_res(callback(v, prefix, locations))
+        mlist.append(norm_res(callback(v, prefix, locations)))
 
     for callback in vel_callbacks(v, 'on_query_completions'):
-        norm_res(callback(prefix, locations))
+        mlist.append(norm_res(callback(prefix, locations)))
 
-    if not completion_lists:
-        completion_lists = [sublime.CompletionList([])]
-
-    mlist = MultiCompletionList(len(completion_lists), view_id, req_id)
-    for cl in completion_lists:
-        cl._set_target(mlist)
+    mlist.ready()
 
 
 def on_hover(view_id, point, hover_zone):
@@ -1245,10 +1182,12 @@ class CommandInputHandler:
         """
         return ""
 
-    def validate(self, text: str) -> bool:
+    def validate(self, text: str, event: Optional[Event] = None) -> bool:
         """
         Called whenever the user presses enter in the text entry box.
         Return :py:`False` to disallow the current value.
+
+        :param event: Only passed when `want_event` returns ``True``.
         """
         return True
 
@@ -1258,10 +1197,12 @@ class CommandInputHandler:
         backspace or escape.
         """
 
-    def confirm(self, text: str):
+    def confirm(self, text: str, event: Optional[Event] = None):
         """
         Called when the input is accepted, after the user has pressed enter and
         the text has been validated.
+
+        :param event: Only passed when `want_event` returns ``True``.
         """
 
     def next_input(self, args) -> Optional[CommandInputHandler]:
@@ -1432,6 +1373,7 @@ class ListInputHandler(CommandInputHandler):
 
         props = {
             "initial_text": self.initial_text(),
+            "initial_selection": self.initial_selection(),
             "placeholder_text": self.placeholder(),
             "selected": selected_item_index,
             "type": "list",
@@ -1647,7 +1589,7 @@ class WindowCommand(Command):
     def __init__(self, window):
         """ :meta private: """
 
-        self.window: Window = window
+        self.window: sublime.Window = window
         """ The `Window` this command is attached to. """
 
     def run_(self, edit_token, args):
@@ -1685,7 +1627,7 @@ class TextCommand(Command):
     def __init__(self, view):
         """ :meta private: """
 
-        self.view: View = view
+        self.view: sublime.View = view
         """ The `View` this command is attached to. """
 
     def run_(self, edit_token, args):
@@ -1810,22 +1752,23 @@ class EventListener:
 
     .. method:: on_pre_move(view: View)
 
-        Called right before a view is moved between two windows, passed the View
-        object.
+        Called right before a view is moved between two windows or within a
+        window. Passed the View object.
 
         .. since:: 4050
 
     .. method:: on_post_move(view: View)
 
-        Called right after a view is moved between two windows, passed the View
-        object.
+        Called right after a view is moved between two windows or within a
+        window. Passed the View object.
 
         .. since:: 4050
 
     .. method:: on_post_move_async(view: View)
 
-        Called right after a view is moved between two windows, passed the View
-        object. Runs in a separate thread, and does not block the application.
+        Called right after a view is moved between two windows or within a
+        window. Passed the View object. Runs in a separate thread, and does not
+        block the application.
 
         .. since:: 4050
 
@@ -1905,7 +1848,8 @@ class EventListener:
         :param hover_zone:
             Which element in Sublime Text the mouse has hovered over.
 
-    .. method:: on_query_context(view: View, key: str, operator: QueryOperator, operand: str, match_all: bool) -> Optional[bool]
+    .. method:: on_query_context(\
+            view: View, key: str, operator: QueryOperator, operand: str, match_all: bool) -> Optional[bool]
 
         Called when determining to trigger a key binding with the given context
         key. If the plugin knows how to respond to the context, it should
@@ -1928,7 +1872,8 @@ class EventListener:
             either does or doesn't match. If the context is unknown return
             ``None``.
 
-    .. method:: on_query_completions(view: View, prefix: str, locations: List[Point]) -> Union[None, List[CompletionValue], Tuple[List[CompletionValue], AutoCompleteFlags], CompletionList]
+    .. method:: on_query_completions(view: View, prefix: str, locations: List[Point]) -> Union[\
+            None, List[CompletionValue], Tuple[List[CompletionValue], AutoCompleteFlags], CompletionList]
 
         Called whenever completions are to be presented to the user.
 
@@ -1942,14 +1887,23 @@ class EventListener:
 
     .. method:: on_text_command(view: View, command_name: str, args: CommandArgs) -> (str, CommandArgs)
 
-        Called when a text command is issued. The listener may return a (command, arguments) tuple to rewrite the command, or None to run the command unmodified.
+        Called when a text command is issued. The listener may return a
+        (command, arguments) tuple to rewrite the command, or ``None`` to run
+        the command unmodified.
 
     .. method:: on_window_command(window: Window, command_name: str, args: CommandArgs) -> (str, CommandArgs)
 
-        Called when a window command is issued. The listener may return a (command, arguments) tuple to rewrite the command, or None to run the command unmodified.
+        Called when a window command is issued. The listener may return a
+        (command, arguments) tuple to rewrite the command, or ``None`` to run
+        the command unmodified.
 
-    .. method:: on_post_text_command(view: View, command_name: str, args: CommandArgs)    Called after a text command has been executed.
-    .. method:: on_post_window_command(window: Window, command_name: str, args: CommandArgs)      Called after a window command has been executed.
+    .. method:: on_post_text_command(view: View, command_name: str, args: CommandArgs)
+
+        Called after a text command has been executed.
+
+    .. method:: on_post_window_command(window: Window, command_name: str, args: CommandArgs)
+
+        Called after a window command has been executed.
 
     .. method:: on_new_window(window: Window)
 
@@ -1959,7 +1913,8 @@ class EventListener:
 
     .. method:: on_new_window_async(window: Window)
 
-        Called when a window is created, passed the Window object. Runs in a separate thread, and does not block the application.
+        Called when a window is created, passed the Window object. Runs in a
+        separate thread, and does not block the application.
 
         .. since:: 4050
 
@@ -1977,7 +1932,8 @@ class EventListener:
 
     .. method:: on_new_project_async(window: Window)
 
-        Called right after a new project is created, passed the Window object. Runs in a separate thread, and does not block the application.
+        Called right after a new project is created, passed the Window object.
+        Runs in a separate thread, and does not block the application.
 
         .. since:: 4050
 
@@ -1989,7 +1945,8 @@ class EventListener:
 
     .. method:: on_load_project_async(window: Window)
 
-        Called right after a project is loaded, passed the Window object. Runs in a separate thread, and does not block the application.
+        Called right after a project is loaded, passed the Window object. Runs
+        in a separate thread, and does not block the application.
 
         .. since:: 4050
 
@@ -2007,7 +1964,8 @@ class EventListener:
 
     .. method:: on_post_save_project_async(window: Window)
 
-        Called right after a project is saved, passed the Window object. Runs in a separate thread, and does not block the application.
+        Called right after a project is saved, passed the Window object. Runs in
+        a separate thread, and does not block the application.
 
         .. since:: 4050
 
@@ -2064,13 +2022,15 @@ class ViewEventListener:
 
     .. method:: on_pre_move()
 
-        Called right before a view is moved between two windows.
+        Called right before a view is moved between two windows or within a
+        window.
 
         .. since:: 4050
 
     .. method:: on_post_move()
 
-        Called right after a view is moved between two windows.
+        Called right after a view is moved between two windows or within a
+        window.
 
         .. since:: 4050
 
@@ -2136,7 +2096,8 @@ class ViewEventListener:
 
     .. method:: on_selection_modified_async()
 
-        Called after the selection has been modified in the view. Runs in a separate thread, and does not block the application.
+        Called after the selection has been modified in the view. Runs in a
+        separate thread, and does not block the application.
 
     .. method:: on_activated()
 
@@ -2144,7 +2105,8 @@ class ViewEventListener:
 
     .. method:: on_activated_async()
 
-        Called when the view gains input focus. Runs in a separate thread, and does not block the application.
+        Called when the view gains input focus. Runs in a separate thread, and
+        does not block the application.
 
     .. method:: on_deactivated()
 
@@ -2152,7 +2114,8 @@ class ViewEventListener:
 
     .. method:: on_deactivated_async()
 
-        Called when the view loses input focus. Runs in a separate thread, and does not block the application.
+        Called when the view loses input focus. Runs in a separate thread, and
+        does not block the application.
 
     .. method:: on_hover(point: Point, hover_zone: HoverZone)
 
@@ -2185,7 +2148,8 @@ class ViewEventListener:
                   and it either does or doesn't match. If the context is unknown
                   return ``None``.
 
-    .. method:: on_query_completions(prefix: str, locations: List[Point]) -> Union[None, List[CompletionValue], Tuple[List[CompletionValue], AutoCompleteFlags], CompletionList]
+    .. method:: on_query_completions(prefix: str, locations: List[Point]) -> Union[\
+            None, List[CompletionValue], Tuple[List[CompletionValue], AutoCompleteFlags], CompletionList]
 
         Called whenever completions are to be presented to the user.
 
@@ -2199,7 +2163,9 @@ class ViewEventListener:
 
     .. method:: on_text_command(command_name: str, args: CommandArgs) -> Tuple[str, CommandArgs]
 
-        Called when a text command is issued. The listener may return a ``(command, arguments)`` tuple to rewrite the command, or ``None`` to run the command unmodified.
+        Called when a text command is issued. The listener may return a
+        `` (command, arguments)`` tuple to rewrite the command, or ``None`` to
+        run the command unmodified.
 
         .. since:: 3155
 
